import { interval, map, Observable, Observer, Subject } from "rxjs"
import { v4 as uuidv4 } from 'uuid'
import { FisMessage, TransportManager } from "../transport/transport.manager"
import { ReceiverRetransmissionProfile, TransportManagerEventNotification } from "../interface/ITransport.interface"
import { RetransmissionService } from "../utils/retransmission.service"

// need to instantiate to grab both incoming and outgoing

class Application {
    service!: Blue
    notification = new Subject<any>()
    constructor(service: Blue) {
        this.service = service

        // this one will directly send over 
        // this.subscribeNotification().subscribe(notification => this.service.transportManager.send(notification))
        this.notify()
    }

    subscribeNotification(): Observable<FisMessage> {
        return this.notification.asObservable()
    }

    notify() {
        // logic here
        let output = interval(1000)
        output.pipe(map(
            value => {
                let message: FisMessage = {
                    header: {
                        messageID: uuidv4(),
                        messageName: `NotificationMessage`
                    },
                    data: `Notification Message ${value}`
                }
                return message
            }
        )).subscribe(this.notification)
    }

    /* EXTRA. Appplication acting as a receiver */
    processRequest(request: FisMessage): Observable<FisMessage> {
        return new Observable((response) => {
            // process responses. Should be called by Blue after receiving a request
        })
    }
    // This two only exist at this level. 
    send(message: FisMessage): Observable<FisMessage> {
        return new Observable((response) => {
            // logic here
        })
    }
}

/* At the highest level, blue will also need to know who the receiver is so that it can establsih said one to one relationsuhip */
class Blue {
    transportManager!: TransportManager
    receiverProfiles: ReceiverRetransmissionProfile[] = []
    event: Subject<TransportManagerEventNotification> = new Subject()
    constructor(transportManager: TransportManager) {
        this.transportManager = transportManager
        this.subscribeForTransportManagerEvent()
        this.manageReceiverCommunication()
    }

    private subscribeForTransportManagerEvent(): void {
        this.transportManager.getTransportManagerEventNotification().subscribe({
            next: (notification: TransportManagerEventNotification) => {
                this.event.next(notification)
            }
        })
    }

    private manageReceiverCommunication(): void {
        this.event.subscribe((notification: TransportManagerEventNotification) => {
            if (notification.event == 'New Client Connection') {
                // first thing first, it will check if it's a previously connected receiver
                this.checkIfPreviouslyConnected(notification.data?.receiverID as string) // but this id is assigned at connection though So that means the receiver must self identify
                let adjustedReceiverProfile: ReceiverRetransmissionProfile = notification.data?.payload
                adjustedReceiverProfile.buffer = new RetransmissionService()
                // retransmission to be activated. Now it's just assigning it there, but idling.
                this.receiverProfiles.push(adjustedReceiverProfile)
                console.log(`BLUE: Added new Receiver Profile ${notification.data?.receiverID} to record. `)
            }
            if (notification.event == 'Client Disconnected') {
                console.log(`BLUE: Client ${notification.data?.receiverID} disconnected. Buffering....`)
                let receiverProfile: ReceiverRetransmissionProfile | undefined = this.receiverProfiles.find(obj => obj.uuid === notification.data?.receiverID)
                if (receiverProfile) {
                    // then notifiy retransmission to stop releasing the buffer 
                }
            }
            if (notification.event == 'New Client Request') {
                this.manageReceiverRequest(notification.data?.payload)
                console.log(`BLUE: Client request ${notification.data?.payload?.header?.messageID ?? `messageID undefined`} received.`)
            }
        })
    }

    private manageReceiverRequest(request: FisMessage): void {

    }

    private checkIfPreviouslyConnected(receiverID: string): boolean {
        return this.receiverProfiles.some(obj => obj.uuid === receiverID)
    }


}


const application = new Application(new Blue(new TransportManager(undefined, 'http://localhost:3000')))

/* Transmitter acting as a receiver */
// logic here